// Package apims wraps (M)ob(sql) exposing mobsql's underlying status and
// load functions.
package apims

import (
	"errors"
	"fmt"
	"git.sr.ht/~mil/mobroute/api/apims/dbschemaextra"
	"git.sr.ht/~mil/mobroute/util/utillog"
	"git.sr.ht/~mil/mobsql"
	"github.com/jmoiron/sqlx"
)

type SourcesetFilter = mobsql.SourcesetFilter
type RuntimeConfig = mobsql.RuntimeConfig
type SourceStatus = mobsql.SourceStatus
type MobsqlRuntime = mobsql.MobsqlRuntime

var MSSourcesetLoad = mobsql.SourcesetLoad
var MSSourcesetCompute = mobsql.SourcesetCompute
var MSSourcesetStatus = mobsql.SourcesetStatus
var MSSourcesetPurge = mobsql.SourcesetPurge

var MSPurgeTablesOptionAll = mobsql.PurgeTablesOptionAll
var MSPurgeTablesOptionGTFS = mobsql.PurgeTablesOptionGTFS
var MSPurgeTablesOptionComputed = mobsql.PurgeTablesOptionComputed

type MSParams struct {
	LoadFilter *SourcesetFilter `yaml:"filter" json:"filter"`
	LoadConfig *RuntimeConfig   `yaml:"config" json:"config"`
}

func MSInitializeRuntime(params *MSParams) (*mobsql.MobsqlRuntime, []int, error) {
	if params.LoadConfig == nil {
		params.LoadConfig = &mobsql.RuntimeConfig{}
	}
	if params.LoadConfig.SchemaExtra == nil {
		params.LoadConfig.SchemaExtra = &dbschemaextra.MobsqlSchemaExtra
	}

	if mobsqlRuntime, err := mobsql.InitializeRuntime(params.LoadConfig); err != nil {
		return nil, nil, fmt.Errorf("Error initializing mobsql runtime: %v", err)
	} else if sourcesetIDs, err := mobsql.FilterToSourceset(mobsqlRuntime, params.LoadFilter); err != nil {
		return nil, nil, fmt.Errorf("Error translating mobsql filter to sourceset: %v", err)
	} else {
		return mobsqlRuntime, sourcesetIDs, nil
	}
	return nil, nil, errors.New("Impossible error creating mobsql runtime")
}

func MSDBOpenRO(params *MSParams) (*sqlx.DB, error) {
	var (
		sqliteCustomDriver = "sqlite3_mobroute_extensions"
		sqliteDB           = fmt.Sprintf("file:%s?mode=ro", dbPath(params.LoadConfig.SQLiteDBPath))
	)
	registerCustomSQLiteDriver(sqliteCustomDriver)
	utillog.Infof("Opening DB in RO mode: %s", sqliteDB)
	if s, err := sqlx.Open(sqliteCustomDriver, sqliteDB); err != nil {
		return nil, err
	} else if _, err = s.Exec(dbschemaextra.MobsqlSchemaExtra.InitExec); err != nil {
		return nil, fmt.Errorf("Error setting pragma statements: %v", err)
	} else {
		return s, nil
	}
}

func MSLoadComputeOpenRODB(mobsqlParams *MSParams) (*sqlx.DB, []int, error) {
	// (1) Mobsql load, compute, and extract mdbids for given filter
	// TODO: should close RW DB from mobsql here created via mobsql runtime
	// as we directly open DB as RO after initialization load/compute logic
	if mobsqlRuntime, mdbids, err := MSInitializeRuntime(mobsqlParams); err != nil {
		return nil, nil, fmt.Errorf("Failed to initialize Mobsql Runtime: %v", err)
	} else if _, err := MSSourcesetLoad(mobsqlRuntime, mdbids); err != nil {
		return nil, nil, fmt.Errorf("Failed to load GTFS via Mobsql: %v", err)
	} else if _, err := MSSourcesetCompute(mobsqlRuntime, mdbids); err != nil {
		return nil, nil, fmt.Errorf("Failed to compute GTFS-dervied tables via Mobsql: %v", err)
	} else if err := mobsqlRuntime.DB.Close(); err != nil {
		return nil, nil, fmt.Errorf("Failed to close mobsql runtime DB: %v", err)
	} else if rodb, err := MSDBOpenRO(mobsqlParams); err != nil {
		return nil, nil, fmt.Errorf("Failed to open SQLite database: %v", err)
	} else {
		return rodb, mdbids, nil
	}
}
