package apiroute

import (
	"fmt"
	"git.sr.ht/~mil/mobroute/api/apims"
	"git.sr.ht/~mil/mobroute/api/apiroute/calcwalk"
	"git.sr.ht/~mil/mobroute/api/apiroute/diagnosticst"
	"git.sr.ht/~mil/mobroute/csa"
	"git.sr.ht/~mil/mobroute/dbquery"
	"git.sr.ht/~mil/mobroute/dbt"
	"git.sr.ht/~mil/mobroute/util/utilfuncs"
	"git.sr.ht/~mil/mobroute/util/utillog"
	"github.com/jmoiron/sqlx"
	"time"
)

// Route performs the routing logic, extracting from the DB the required
// data for CSA, running the CSA algorithm, going back to the DB to pull a few
// extra fields for presentation, and finally formatting the data into route
// legs for easy consumption:
//
// Essentially a 4 step process:
//
//  1. CSA Inputs Dataextract:
//     Extracts necessary data from the SQLite database, 4 inputs
//     extracted: connectionss, transfers, walksfrom, and walksto
//
//  2. CSA Execution:
//     Runs the CSA algorithm (returns []*dbt.Connection)
//
//  3. CSA Outputs Verbosity Dataextract:
//     Result of CSA algorithm is fitered []*dbt.Connection, which is a minimal
//     datastructure with just information needed for CSA - so in this
//     step we go back to the database to extract []*dbt.ConnectionVerbose
//     for the result of the CSA alg's connections.
//
//  4. Leg Formatting:
//     Formatting the []*dbt.ConnectionVerbose CSA results into a nice pretty
//     format showing each part of the trip:
//     - Leg for walking to input stop
//     - (n) trips (with each []*dbt.Connection nested within)
//     - (n) transfers between each trip
//     - Leg for walking from output stop to destination
func Route(paramsGiven *RouteParams) (*RouteResponse, error) {
	var (
		params               = paramsValueify(paramsGiven)
		timeYYYYMMDD         = utilfuncs.DayYYYYMMDD(params.Time)
		_, timeUTCOffsetSecs = params.Time.Zone()
		timeSecondsFrom      = timeUTCOffsetSecs + int(utilfuncs.DaySeconds(params.Time))
		timeSecondsTo        = timeSecondsFrom + int(*params.MaxTripSeconds)
		maxKmWalkSeconds     = calcwalk.WalkKmInSecondsGivenKmHr(*params.MaxWalkSeconds, *params.WalkspeedKmHr)
		timer                = Timer{lastTime: time.Now()}
		diagnostics          = diagnosticst.Diagnostics{}

		err                      error
		db                       *sqlx.DB
		mdbids                   []int
		memConns                 *[]dbt.Connection
		memConnsverbose          *[]dbt.ConnectionVerbose
		memTransfers             *[]dbt.Transfer
		memWalksfrom, memWalksto *[]dbt.StopWalk
		csaConns                 []*dbt.Connection
	)

	// 0. Log route params for reproducibility
	utillog.Infoj("Running route API against input", paramsGiven)

	// 1. Initialize Mobsql
	//   - Initialize Mobsql's Runtime
	//   - Translate mobsql params into mdbids
	//   - Runs Mobsql's SourcesetLoad to load raw GTFS tables
	//   - Runs Mobsql's SourcesetCompute to compute GTFS-derived tables
	//   - Opens DB as RO once all previous steps prepared and DB ready for calcs
	if db, mdbids, err = apims.MSLoadComputeOpenRODB(params.MSParams); err == nil {
		diagnostics.PerfSLoadcompute = timer.Snapshot()
	} else {
		return nil, fmt.Errorf("Failed to prepare DB via mobsql: %v", err)
	}

	// (2) Sanity checks
	if err = routeRequestSanityChecks(db, mdbids, &params); err != nil {
		return nil, fmt.Errorf("Route request failed sanity check: %v", err)
	}

	// (3.1) Memload connections
	if memConns, err = dbquery.LoadConnections(
		db, mdbids, timeYYYYMMDD, timeSecondsFrom, timeSecondsTo,
	); err != nil {
		return nil, fmt.Errorf("Failed to load connections from DB: %v", err)
	} else if len(*memConns) == 0 {
		return nil, fmt.Errorf(
			"Empty connections set for MDBIDs %v, date %d from UTC seconds %d -> %d",
			mdbids, timeYYYYMMDD, timeSecondsFrom, timeSecondsTo,
		)
	} else {
		diagnostics.PerfSMemloadConns = timer.Snapshot()
		diagnostics.PerfNMemConns = len(*memConns)
	}

	// (3.2) Memload transfers
	if memTransfers, err = dbquery.LoadTransfers(
		db,
		mdbids,
		*params.MinTransferSeconds, *params.MaxTransferSeconds,
		*params.WalkspeedKmHr,
		*params.TransferCategories,
	); err != nil {
		return nil, fmt.Errorf("Failed to load transfers from DB: %v", err)
	} else {
		diagnostics.PerfSMemloadTransfers = timer.Snapshot()
		diagnostics.PerfNMemTransfers = len(*memTransfers)
	}

	// (3.3) Memload walksfrom
	if memWalksfrom, err = dbquery.LoadStopWalks(db, mdbids, *params.From, maxKmWalkSeconds); err != nil {
		return nil, fmt.Errorf("Failed to load walksfrom from DB: %v", err)
	} else if nonNilWalksN := calcwalk.CountNonNilWalks(memWalksfrom); nonNilWalksN == 0 {
		return nil, fmt.Errorf("Failed to determine >0 stopwalks to from dest: %v within max walk km: %0.2f", *params.From, maxKmWalkSeconds)
	} else {
		diagnostics.PerfSMemloadWalksfrom = timer.Snapshot()
		diagnostics.PerfNMemWalksfrom = len(*memWalksfrom)
		diagnostics.PerfNMemWalksfromNonnil = nonNilWalksN
	}

	// (3.4) Memload walksto
	if memWalksto, err = dbquery.LoadStopWalks(db, mdbids, *params.To, maxKmWalkSeconds); err != nil {
		return nil, fmt.Errorf("Failed to load walksto from DB: %v", err)
	} else if nonNilWalksN := calcwalk.CountNonNilWalks(memWalksto); nonNilWalksN == 0 {
		return nil, fmt.Errorf("Failed to determine >0 stopwalks to to dest: %v within max walk km: %0.2f", *params.To, maxKmWalkSeconds)
	} else {
		diagnostics.PerfSMemloadWalksto = timer.Snapshot()
		diagnostics.PerfNMemWalksto = len(*memWalksto)
		diagnostics.PerfNMemWalkstoNonnil = nonNilWalksN
	}

	// (4) Run CSA
	if csaConns, err = csa.ExecuteCSA( // (2)
		memConns,
		memTransfers,
		calcwalk.WalksToMapTimeSecs(memWalksfrom, *params.WalkspeedKmHr),
		calcwalk.WalksToMapTimeSecs(memWalksto, *params.WalkspeedKmHr),
		int(timeSecondsFrom),
		*params.MaxTransferSeconds,
		*params.MaxWalkSeconds,
		*params.MaxNTransfers,
	); err != nil {
		return nil, fmt.Errorf("Failed to run CSA: %v", err)
	} else {
		diagnostics.PerfSRoute = timer.Snapshot()
	}

	// (5)  Verbose connections
	if memConnsverbose, err = dbquery.LoadConnectionsVerbose( // (3)
		db, mdbids, timeYYYYMMDD, timeSecondsFrom, timeSecondsTo,
		csa.ConnIDs(csaConns),
	); err != nil {
		return nil, err
	} else {
		diagnostics.PerfSMemloadConnsverb = timer.Snapshot()
	}

	// (6)
	db.Close()
	return formatRouteResponse(&diagnostics, memConnsverbose, memTransfers, &params), nil
}

func routeRequestSanityChecks(db *sqlx.DB, mdbids []int, dbParams *RouteParams) error {
	if err := dbquery.QueryBoundsMatch(db, mdbids, *dbParams.From); err != nil {
		return fmt.Errorf("From %v failed to match any bounds: %v", dbParams.From, err)
	} else if err := dbquery.QueryBoundsMatch(db, mdbids, *dbParams.To); err != nil {
		return fmt.Errorf("To %v failed to match any bounds: %v", dbParams.To, err)
	} else if *dbParams.MaxTransferSeconds < *dbParams.MinTransferSeconds {
		return fmt.Errorf(
			"Failed condition: %d max_transfer_seconds < %d min_transfer_seconds",
			*dbParams.MaxTransferSeconds, *dbParams.MinTransferSeconds,
		)
	} else {
		return nil
	}
}

func paramsValueify(b *RouteParams) RouteParams {
	return RouteParams{
		MSParams: coalesce(b.MSParams, &apims.MSParams{
			LoadFilter: &apims.SourcesetFilter{Mdbid: []int{516}},
		}),
		From:               coalesce(b.From, &utilfuncs.Coord{Lat: 40.70940, Lon: -74.00537}),
		To:                 coalesce(b.To, &utilfuncs.Coord{Lat: 40.72879, Lon: -73.95215}),
		Time:               coalesce(b.Time, ptr(time.Now())),
		MaxWalkSeconds:     coalesce(b.MaxWalkSeconds, ptr(uint(20*60))),     // 20min
		MaxTripSeconds:     coalesce(b.MaxTripSeconds, ptr(uint(240*60))),    // 240min (4hrs)
		MinTransferSeconds: coalesce(b.MinTransferSeconds, ptr(uint(3*60))),  // 3min
		MaxTransferSeconds: coalesce(b.MaxTransferSeconds, ptr(uint(40*60))), // 40min
		WalkspeedKmHr:      coalesce(b.WalkspeedKmHr, ptr(3.5)),
		TransferCategories: coalesce(b.TransferCategories, ptr([]string{"f", "i"})),
		MaxNTransfers:      coalesce(b.MaxNTransfers, ptr(uint(20))),
		OutputFormats: coalesce(b.OutputFormats, ptr([]string{
			"legs", "diagnostics", "mapurl", "request",
		})),
	}
}
