//go:build zipgtfs

package apiroute_test

import (
	"bytes"
	_ "embed"
	"encoding/json"
	"fmt"
	"github.com/gocarina/gocsv"
	"os"
	"path/filepath"
	"strconv"
	"strings"
	"testing"
	"text/template"
	"time"

	"git.sr.ht/~mil/mobroute/api/apims"
	"git.sr.ht/~mil/mobroute/api/apiroute"
	"git.sr.ht/~mil/mobroute/util/utilfuncs"
)

//go:embed resources/test_results_template.html
var testResultsHTMLTemplate string

//go:embed resources/realistic_mobilitydb.csv
var realisticMobilityDBCSV []byte

var (
	enableValidateTransfers bool = true
	enableValidateDurations bool = true
)

// Adding examples:
//
//  1. Upload GTFS archive for MDBID source to https://ci.lrdu.org/tests_gtfs/
//     (note filenames are just automatically generated based MR format, e.g.
//     just copy from ~/.cache/mobroute/gtfs_*)
//
//  2. Regenerate mock MobilityDB CSV:
//     ./scripts/prepare_mockmobdbcsv.sh > api/apiroute_test/resources/realistic_mobilitydb.csv
//
//  3. Then add example below
var realisticTestCases = []struct {
	Mdbids             []int
	TransferCategories []string
	From               [2]float64
	To                 [2]float64
	Time               string
	MaxWalkSeconds     *uint

	ExpectTransfers uint
	ExpectDuration  uint
}{
	{Mdbids: []int{1270}, TransferCategories: []string{"f", "i"}, From: [2]float64{50.012338, 19.88192}, To: [2]float64{50.08785, 20.02387}, Time: "2024-02-02T21:19:40+01:00", ExpectTransfers: 0, ExpectDuration: 4168},
	{Mdbids: []int{516}, TransferCategories: []string{"f", "i"}, From: [2]float64{40.70940, -74.00537}, To: [2]float64{40.72879, -73.95215}, Time: "2024-02-02T21:19:40-05:00", ExpectTransfers: 2, ExpectDuration: 2264},
	{Mdbids: []int{509}, TransferCategories: []string{"f", "i"}, From: [2]float64{39.6317, -74.7995}, To: [2]float64{39.9778, -75.0618}, Time: "2024-02-02T21:19:40-05:00", ExpectTransfers: 0, ExpectDuration: 2843},
	{Mdbids: []int{1088}, TransferCategories: []string{"f", "i"}, From: [2]float64{50.85728, 4.351426}, To: [2]float64{50.83214, 4.350534}, Time: "2024-02-02T21:19:40+01:00", ExpectTransfers: 0, ExpectDuration: 920},
	{Mdbids: []int{1898}, TransferCategories: []string{"f", "i"}, From: [2]float64{47.07199, 2.40154}, To: [2]float64{47.10811, 2.41974}, Time: "2024-02-02T21:19:40+01:00", ExpectTransfers: 0, ExpectDuration: 2207},
	{Mdbids: []int{1032}, TransferCategories: []string{"f", "i"}, From: [2]float64{38.7376, -9.1725}, To: [2]float64{38.72329, -9.14646}, Time: "2024-02-02T21:19:40Z", ExpectTransfers: 0, ExpectDuration: 2329},
	{Mdbids: []int{767}, TransferCategories: []string{"f", "i", "g"}, From: [2]float64{50.0767, 14.4732}, To: [2]float64{50.1060, 14.3690}, Time: "2024-03-30T21:19:40+01:00", MaxWalkSeconds: ptr(uint(700)), ExpectTransfers: 1, ExpectDuration: 2733},
	{Mdbids: []int{524}, TransferCategories: []string{"f", "i"}, From: [2]float64{40.9599, -73.8202}, To: [2]float64{41.3942, -73.6192}, Time: "2024-03-30T21:19:40-04:00", ExpectTransfers: 1, ExpectDuration: 4700},
	{Mdbids: []int{660}, TransferCategories: []string{"f", "i"}, From: [2]float64{-35.0435, 138.8789}, To: [2]float64{-34.919223, 138.575316}, Time: "2024-03-30T13:53:35+10:30", ExpectTransfers: 2, ExpectDuration: 6368},
	{Mdbids: []int{1828}, TransferCategories: []string{"f", "i"}, From: [2]float64{38.2094, 27.3072}, To: [2]float64{38.4222, 27.157}, Time: "2024-03-30T13:59:30+03:00", ExpectTransfers: 0, ExpectDuration: 3034},
	{Mdbids: []int{509, 516}, TransferCategories: []string{"f", "i", "g"}, From: [2]float64{40.64809, -73.93799}, To: [2]float64{40.82732, -74.12064}, Time: "2024-02-02T21:19:40-05:00", ExpectTransfers: 2, ExpectDuration: 8069},
} //

func Test_ApiRouteZIPGTFS(t *testing.T) {
	var (
		saveDir       = "../../test_results"
		mockMDBCSV, _ = filepath.Abs(saveDir + "/mockmdb.csv")
		fileSqliteDB  *os.File
		err           error
	)

	if fileSqliteDB, err = os.CreateTemp("", "mobroute_test_sqlite_*.db"); err != nil {
		t.Fatalf("Failed to create mock SQLite DB: %v", err)
	} else if err := os.MkdirAll(saveDir, os.ModePerm); err != nil {
		t.Fatalf("Error creating filepath; %v", err)
	} else if err := os.WriteFile(mockMDBCSV, realisticMobilityDBCSV, 0644); err != nil {
		t.Fatalf("Failed to get executable path: %v", err)
	}
	defer os.Remove(fileSqliteDB.Name())

	results := []*apiroute.RouteResponse{}
	for i, testCase := range realisticTestCases {
		var (
			saveFile = fmt.Sprintf("%s/json/%s.json", saveDir, arrayIntToKey(testCase.Mdbids))
		)
		if response, err := apiroute.Route(
			&apiroute.RouteParams{
				MSParams: &apims.MSParams{
					LoadConfig: &apims.RuntimeConfig{
						MobilityDBCatalogCSVURI: fmt.Sprintf("file://%s", mockMDBCSV),
						SQLiteDBPath:            fileSqliteDB.Name(),
						AllowFileURIs:           true,
						LogInfo:                 true,
						LogWarn:                 true,
						LogDebug:                true,
					},
					LoadFilter: &apims.SourcesetFilter{Mdbid: testCase.Mdbids},
				},
				From:               &utilfuncs.Coord{Lat: testCase.From[0], Lon: testCase.From[1]},
				To:                 &utilfuncs.Coord{Lat: testCase.To[0], Lon: testCase.To[1]},
				TransferCategories: &testCase.TransferCategories,
				Time:               timeParse(t, testCase.Time),
				MaxWalkSeconds:     testCase.MaxWalkSeconds,
			},
		); err != nil {
			t.Fatalf("(%d / %d) Error with calculating route: %v", i, len(realisticTestCases), err)
		} else if enableValidateDurations && testCase.ExpectDuration != response.Diagnostics.RouteinfoDuration {
			t.Fatalf(
				"(%d / %d) Duration expected: %v, got: %v",
				i, len(realisticTestCases), testCase.ExpectDuration, response.Diagnostics.RouteinfoDuration,
			)
		} else if enableValidateTransfers && testCase.ExpectTransfers != response.Diagnostics.RouteinfoTransfers {
			t.Fatalf(
				"(%d / %d) Transfers expected: %v, got: %v",
				i, len(realisticTestCases), testCase.ExpectTransfers, response.Diagnostics.RouteinfoTransfers,
			)
		} else if responseJSONBytes, err := json.MarshalIndent(response, "", "\t"); err != nil {
			t.Fatalf("(%d / %d) Error marshalling JSON: %v", i, len(realisticTestCases), err)
		} else if err := os.MkdirAll(saveDir+"/json", os.ModePerm); err != nil {
			t.Fatalf("(%d / %d) Error creating filepath; %v", i, len(realisticTestCases), err)
		} else if err := os.WriteFile(saveFile, responseJSONBytes, 0644); err != nil {
			t.Fatalf("(%d / %d) Error writing file: %v", i, len(realisticTestCases), err)
		} else {
			results = append(results, response)
		}
	}

	if resultStr, err := testsSummaryHTMLPageCreate(mockMDBCSV, results); err != nil {
		t.Fatalf("Error creating summary page: %v", err)
	} else if err := os.WriteFile(saveDir+"/index.html", []byte(resultStr), 0644); err != nil {
		t.Fatalf("Error writing file: %v", err)
	}
}

func timeParse(t *testing.T, str string) *time.Time {
	var (
		err        error
		parsedTime time.Time
	)
	if parsedTime, err = time.Parse(time.RFC3339, str); err != nil {
		t.Fatalf("Fail to parse time: %v", err)
	}
	return &parsedTime
}

func mdbMetadataMap(fileMobdbCSVPath string) (map[string]map[string]string, error) {
	type MDBEntry struct {
		Mdbid        int    `csv:"mdb_source_id"`
		Country      string `csv:"location.country_code"`
		Subdivision  string `csv:"location.subdivision_name"`
		Municipality string `csv:"location.municipality"`
		Provider     string `csv:"provider"`
		Name         string `csv:"name"`
	}
	entries := []*MDBEntry{}

	if fileMobdbCSV, err := os.OpenFile(fileMobdbCSVPath, os.O_RDWR, os.ModePerm); err != nil {
		return nil, err
	} else if err := gocsv.UnmarshalFile(fileMobdbCSV, &entries); err != nil {
		return nil, err
	} else {
		returnMap := map[string]map[string]string{}

		// Single items
		for _, parsed := range entries {
			returnMap[arrayIntToKey([]int{parsed.Mdbid})] = map[string]string{
				"Mdbids":       arrayIntToKey([]int{parsed.Mdbid}),
				"Country":      coalesce([]string{parsed.Country, "Unset"}),
				"Municipality": coalesce([]string{parsed.Municipality, parsed.Subdivision, "Unset"}),
				"Name": func() string {
					if parsed.Provider != "" && parsed.Name != "" {
						return fmt.Sprintf("%s: %s", parsed.Provider, parsed.Name)
					} else if parsed.Provider != "" {
						return parsed.Provider
					} else {
						return "Unset"
					}
				}(),
			}
		}

		// Multisource (only 2 combos supported for now in test fmting)
		for _, parsedA := range entries {
			for _, parsedB := range entries {
				var (
					aMdbid   = strconv.Itoa(parsedA.Mdbid)
					aCountry = returnMap[aMdbid]["Country"]
					aMuni    = returnMap[aMdbid]["Municipality"]
					aName    = returnMap[aMdbid]["Name"]

					bMdbid   = strconv.Itoa(parsedB.Mdbid)
					bCountry = returnMap[bMdbid]["Country"]
					bMuni    = returnMap[bMdbid]["Municipality"]
					bName    = returnMap[bMdbid]["Name"]
				)
				returnMap[arrayIntToKey([]int{parsedA.Mdbid, parsedB.Mdbid})] = map[string]string{
					"Mdbids":       fmt.Sprintf("%s + %s", aMdbid, bMdbid),
					"Country":      fmt.Sprintf("%s + %s", aCountry, bCountry),
					"Municipality": fmt.Sprintf("%s + %s", aMuni, bMuni),
					"Name":         fmt.Sprintf("%s + %s", aName, bName),
				}
			}
		}
		return returnMap, nil
	}
	return nil, nil
}

func testsSummaryHTMLPageCreate(fileMobdbCSVPath string, results []*apiroute.RouteResponse) (string, error) {
	var (
		htmlTemplate = testResultsHTMLTemplate
		resultBuffer bytes.Buffer
	)
	if metadataMap, err := mdbMetadataMap(fileMobdbCSVPath); err != nil {
		return "", err
	} else if tmpl, err := template.New("summary_tmpl").Funcs(template.FuncMap{
		"arrayIntToKey":           arrayIntToKey,
		"jsonify":                 jsonify,
		"makeMRRequestRepeatable": makeMRRequestRepeatable,
	}).Parse(htmlTemplate); err != nil {
		return "", err
	} else if err := tmpl.ExecuteTemplate(&resultBuffer, "summary_tmpl", map[string]any{
		"Responses": results,
		"Metadata":  metadataMap,
	}); err != nil {
		return "", err
	}
	return resultBuffer.String(), nil
}

func arrayIntToKey(ints []int) string {
	return strings.Join(utilfuncs.IntArrayToStringArray(ints), "_")
}

func makeMRRequestRepeatable(p *apiroute.RouteParams) *apiroute.RouteParams {
	// E.g. just hooks the CSV we used (which is uploaded to below location)
	// so that the entire request can just be dumped into
	// ./mobroute -p '{BUF}' to repeat/debug the request locally wholesale
	p.MSParams.LoadConfig.MobilityDBCatalogCSVURI = "http://ci.lrdu.org/tests_results/mockmdb.csv"
	p.MSParams.LoadConfig.SQLiteDBPath = ""
	return p
}

func jsonify(data any) string {
	b, _ := json.Marshal(data)
	return string(b)
}

func coalesce(strs []string) string {
	for _, s := range strs {
		if s != "" {
			return s
		}
	}
	return ""
}

func ptr[T any](a T) *T {
	return &a
}
