// Package main is the CLI interface for mobroute consuming the mobroute API.
// See mobroute -h for usage information.
package main

import (
	"encoding/json"
	"errors"
	"flag"
	"fmt"

	"log"
	"os"
	"runtime/pprof"
	"strings"

	"git.sr.ht/~mil/mobroute"

	"git.sr.ht/~mil/mobroute/util/utilfuncs"
	"git.sr.ht/~mil/mobroute/util/utillog"
	"gopkg.in/yaml.v3"
)

func main() {
	var (
		jsonBytes               []byte
		commandData             any
		err                     error
		flagBaseDebugCategories string
		flagBaseDebugProfile    string
		flagBaseOutputFormat    string

		commandDefaultsRoute = "{load: {filter: {mdbid: [516]}}, from: {lat: 40.70940, lon: -74.00537}, to: {lat: 40.72879, lon: -73.95215}, output_formats: [legs, diagnostics]}"
		commandDefaultsMS    = "{filter: {mdbid: [516]}, config: {}}"

		flagMSSet    = flag.NewFlagSet("ms", flag.ExitOnError)
		flagMSParams = flagMSSet.String("f", commandDefaultsMS, "")

		flagRouteSet    = flag.NewFlagSet("route", flag.ExitOnError)
		flagRouteParams = flagRouteSet.String(
			"p",
			commandDefaultsRoute,
			`YAML Parameters for db subcommand; note JSON is also valid YAML as well

Valid properties:
  - filter: Object based on mobsql's filter specification, see git.sr.ht/~mil/mobsql (default: {mdbid: [516]})
  - from: Object with two float properties, lat & lon (default: {lat: 40.70940, lon: -74.00537})
  - to: Object with two float properties, lat & lon (default: {lat: 40.72879, lon: -73.95215})
  - time: Datetime to depart from trip, if unset defaults to current time / now (default: null)
  - max_walk_seconds: Maximum seconds walking distance on start/end of trip to get to/from origin/destination stops (default: 1200)
  - max_trip_seconds: Maximum seconds the entire route/trip can take (default: 14400)
  - min_transfer_seconds: Minimum seconds each transfer in the route can take (default: 180)
  - max_transfer_seconds: Maximum seconds each transfer in the route can take (default: 2400)
  - max_n_transfers: Maximum number of transfers a route can contain (default: 20)
  - walkspeed_km_hr: Ratio of kilometers per hour the user walk (default: 4.5)
  - transfer_categories: Array of transfers categories to load - "f" for feed, "i" for implicit, "g" for generated (default: ["f", "i"])
  - cache_dir: Directory to store sqlite db, defaults to ~/.cache/mobroute if null (default: null)
  - output_formats: Array of formats for the output may include - "legs", "geojson", "diagnostics", "mapurl", "request" (default: ["legs"])
	`)
	)

	// Same usage for all subcommands
	for _, f := range []*flag.FlagSet{flagRouteSet} {
		f.StringVar(
			&flagBaseDebugCategories, "d", "i",
			"Verbose debug categories: (i) Information (w) Warn (d) Debug",
		)
		f.StringVar(
			&flagBaseDebugProfile, "dprof", "",
			"Write debug pprof CPU profile to file",
		)
		f.StringVar(
			&flagBaseOutputFormat, "format", "json",
			"Output format, currently only json supported",
		)
	}

	validSubcommands := []string{
		"route",
		"version",
		"msload",
		"msstatus",
		"mscompute",
		"mspurgegtfs",
		"mspurgecomputed",
		"mspurgeall",
	}

	if len(os.Args) < 2 || !utilfuncs.StringInArray(validSubcommands, os.Args[1]) {
		fmt.Println("Provide valid subcommand as first arg, valid subcommands:", validSubcommands)
		os.Exit(1)
	}

	if os.Args[1] == "route" {
		flagRouteSet.Parse(os.Args[2:])
		globalDeferedFn := globalsSetup(flagBaseDebugCategories, flagBaseDebugProfile)
		defer globalDeferedFn()
		commandData, err = cmdRoute(*flagRouteParams)
	} else if utilfuncs.ItemInArray(
		os.Args[1],
		[]string{"msstatus", "msload", "mscompute", "mspurgegtfs", "mspurgecomputed", "mspurgeall"},
	) {
		flagMSSet.Parse(os.Args[2:])
		commandData, err = cmdMS(*flagMSParams, os.Args[1])
	} else if os.Args[1] == "version" {
		commandData = "v0.5.0"
	}

	//_ = format
	if err != nil {
		errorDie(err, false)
	} else if jsonBytes, err = json.MarshalIndent(commandData, "", "  "); err != nil {
		errorDie(err, false)
	} else {
		fmt.Println(string(jsonBytes))
	}
}

func cmdRoute(paramsYAML string) (any, error) {
	var routeParams mobroute.RouteParams
	if err := yaml.Unmarshal([]byte(paramsYAML), &routeParams); err != nil {
		return nil, fmt.Errorf("Error demarshalling YML %s: %v", paramsYAML, err)
	} else {
		return mobroute.Route(&routeParams)
	}
}

func cmdMS(paramsYAML, subcommand string) (any, error) {
	var dbParams mobroute.MSParams
	if err := yaml.Unmarshal([]byte(paramsYAML), &dbParams); err != nil {
		return nil, fmt.Errorf("Error demarshalling YML %s: %v", paramsYAML, err)
	} else if msRuntime, mdbids, err := mobroute.MSInitializeRuntime(&dbParams); err != nil {
		return nil, fmt.Errorf("Error initialize mobsql runtime: %v", err)
	} else if subcommand == "msstatus" {
		return mobroute.MSSourcesetStatus(msRuntime, mdbids)
	} else if subcommand == "msload" {
		return mobroute.MSSourcesetLoad(msRuntime, mdbids)
	} else if subcommand == "mscompute" {
		return mobroute.MSSourcesetCompute(msRuntime, mdbids)
	} else if subcommand == "mspurgegtfs" {
		return mobroute.MSSourcesetPurge(msRuntime, mdbids, mobroute.MSPurgeTablesOptionGTFS)
	} else if subcommand == "mspurgecomputed" {
		return mobroute.MSSourcesetPurge(msRuntime, mdbids, mobroute.MSPurgeTablesOptionComputed)
	} else if subcommand == "mspurgeall" {
		return mobroute.MSSourcesetPurge(msRuntime, mdbids, mobroute.MSPurgeTablesOptionAll)
	} else {
		return nil, errors.New("Unknown subcommand for mobsql")
	}
}

func errorDie(err error, showUsage bool) {
	if showUsage {
		flag.Usage()
	}
	if err != nil {
		utillog.Fatal(err)
	}
	os.Exit(1)
}

func globalsSetup(debugCategories, debugProfile string) func() {
	// Setup debug categories
	if strings.Contains(debugCategories, "i") {
		utillog.LogInfo = true
	}
	if strings.Contains(debugCategories, "w") {
		utillog.LogWarn = true
	}
	if strings.Contains(debugCategories, "d") {
		utillog.LogDebug = true
	}

	// Setup pprof debugging
	if debugProfile != "" {
		if f, err := os.Create(debugProfile); err != nil {
			log.Fatal(err)
		} else {
			pprof.StartCPUProfile(f)
			return pprof.StopCPUProfile
		}
	}

	return func() {}
}
