package dbquery_test

import (
	"fmt"
	"os"
	"reflect"
	"testing"

	"git.sr.ht/~mil/mobroute/dbquery"
	"git.sr.ht/~mil/mobroute/dbt"
	"git.sr.ht/~mil/mobroute/testhelper/dbquerytesthelper"
	"git.sr.ht/~mil/mobroute/util/utilfuncs"
)

func Test_LoadStopWalks(t *testing.T) {
	testCases := []struct {
		Test            string
		MockGTFS        map[int]string
		InputGTFSFile   string
		Sources         []int
		OriginCoord     utilfuncs.Coord
		MaxWalkKM       float64
		ExpectStopWalks []dbt.StopWalk
	}{
		{
			Test:        "Properly determines stopwalks distances based on 'ultimate' stations",
			MockGTFS:    map[int]string{42: "../testhelper/mockgtfs/gtfs_simple_stops.yml"},
			Sources:     []int{42},
			OriginCoord: utilfuncs.Coord{Lat: 1.1, Lon: 0.2},
			MaxWalkKM:   50,
			ExpectStopWalks: []dbt.StopWalk{
				// StopSIDU, StopName, DistKM
				{"42_blankplatform_no_parent", "blankplatform_no_parent", ptr(18.9006)},
				{"42_platform_no_parent", "platform_no_parent", ptr(24.0524)},
				{"42_station_a", "station_a", ptr(27.4599)},
				{"42_station_b", "station_b", ptr(27.8618)},
			},
		},
		{
			Test:        "Returns stops with null distance if they are beyond the given max walk km threshold",
			MockGTFS:    map[int]string{42: "../testhelper/mockgtfs/gtfs_simple_stops.yml"},
			Sources:     []int{42},
			OriginCoord: utilfuncs.Coord{Lat: 1.1, Lon: 0.2},
			MaxWalkKM:   20,
			ExpectStopWalks: []dbt.StopWalk{
				// StopSIDU, StopName, DistKM
				{"42_platform_no_parent", "platform_no_parent", nil},
				{"42_station_a", "station_a", nil},
				{"42_station_b", "station_b", nil},
				{"42_blankplatform_no_parent", "blankplatform_no_parent", ptr(18.9006)},
			},
		},
	}

	for _, testCase := range testCases {
		if sqlxDB, dbPath, err := dbquerytesthelper.PrepareMockLoadComputeAndOpenDB(
			testCase.MockGTFS,
		); err != nil {
			t.Fatalf("Error preparing testcase: %v", err)
		} else if stopwalks, err := dbquery.LoadStopWalks(
			sqlxDB, testCase.Sources, testCase.OriginCoord, testCase.MaxWalkKM,
		); err != nil {
			t.Fatalf("Error: %v: DB: %s", err, dbPath)
		} else if !reflect.DeepEqual(
			MapStopWalksToStrings(testCase.ExpectStopWalks),
			MapStopWalksToStrings(*stopwalks),
		) {
			t.Fatalf(
				"LoadStopWalks Test failed: %s - expected: %v; got: %v; DB: (%s)",
				testCase.Test,
				MapStopWalksToStrings(testCase.ExpectStopWalks),
				MapStopWalksToStrings(*stopwalks),
				dbPath,
			)
		} else {
			os.Remove(dbPath)
		}
	}
}

// MapStopWalksToStrings maps the []StopWalk into an [][]string so that
// values (for ptrs) can be simply compared in test logic
func MapStopWalksToStrings(arr []dbt.StopWalk) [][]string {
	strs := [][]string{}
	for _, sw := range arr {
		distKM := ""
		if sw.DistKM != nil {
			distKM = fmt.Sprintf("%0.4f", *sw.DistKM)
		}
		strs = append(strs, []string{sw.StopSIDU, sw.StopName, distKM})
	}
	return strs
}

func ptr[T any](a T) *T {
	return &a
}
